@@ -24,6 +24,7 @@
 #include <osmocom/mgcp/mgcp_internal.h>
 #include <osmocom/mgcp/osmux.h>
 #include <osmocom/mgcp/mgcp_conn.h>
+#include <osmocom/mgcp/mgcp_endp.h>
 
 static struct osmo_fd osmux_fd;
 
@@ -142,7 +143,7 @@ osmux_handle_alloc(struct mgcp_config *cfg, struct in_addr *addr, int rem_port)
 }
 
 /* Lookup existing handle for a specified address, if the handle can not be
- * foud a the function will automatically allocate one */
+ * found, the function will automatically allocate one */
 static struct osmux_in_handle *
 osmux_handle_lookup(struct mgcp_config *cfg, struct in_addr *addr, int rem_port)
 {
@@ -207,12 +208,18 @@ endpoint_lookup(struct mgcp_config *cfg, int cid,
 		case MGCP_DEST_NET:
 			/* FIXME: Get rid of CONN_ID_XXX! */
 			conn_net = mgcp_conn_get_rtp(endp, CONN_ID_NET);
-			this = &conn_net->end.addr;
+			if (conn_net)
+				this = &conn_net->end.addr;
+			else
+				this = NULL;
 			break;
 		case MGCP_DEST_BTS:
 			/* FIXME: Get rid of CONN_ID_XXX! */
 			conn_bts = mgcp_conn_get_rtp(endp, CONN_ID_BTS);
-			this = &conn_bts->end.addr;
+			if (conn_bts)
+				this = &conn_bts->end.addr;
+			else
+				this = NULL;
 			break;
 		default:
 			/* Should not ever happen */
@@ -222,7 +229,8 @@ endpoint_lookup(struct mgcp_config *cfg, int cid,
 
 		/* FIXME: Get rid of CONN_ID_XXX! */
 		conn_net = mgcp_conn_get_rtp(endp, CONN_ID_NET);
-		if (conn_net->osmux.cid == cid && this->s_addr == from_addr->s_addr)
+		if (conn_net && this && conn_net->osmux.cid == cid
+		    && this->s_addr == from_addr->s_addr)
 			return endp;
 	}
 
@@ -248,8 +256,8 @@ static void scheduled_tx_net_cb(struct msgb *msg, void *data)
 		.sin_port = conn_net->end.rtp_port,
 	};
 
-	conn_bts->end.octets_tx += msg->len;
-	conn_bts->end.packets_tx++;
+	rate_ctr_inc(&conn_bts->rate_ctr_group->ctr[RTP_PACKETS_TX_CTR]);
+	rate_ctr_add(&conn_bts->rate_ctr_group->ctr[RTP_OCTETS_TX_CTR], msg->len);
 
 	/* Send RTP data to NET */
 	/* FIXME: Get rid of conn_bts and conn_net! */
@@ -275,8 +283,8 @@ static void scheduled_tx_bts_cb(struct msgb *msg, void *data)
 		.sin_port = conn_bts->end.rtp_port,
 	};
 
-	conn_net->end.octets_tx += msg->len;
-	conn_net->end.packets_tx++;
+	rate_ctr_inc(&conn_net->rate_ctr_group->ctr[RTP_PACKETS_TX_CTR]);
+	rate_ctr_add(&conn_net->rate_ctr_group->ctr[RTP_OCTETS_TX_CTR], msg->len);	
 
 	/* Send RTP data to BTS */
 	/* FIXME: Get rid of conn_bts and conn_net! */
@@ -314,11 +322,10 @@ int osmux_read_from_bsc_nat_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	struct msgb *msg;
 	struct osmux_hdr *osmuxh;
-	struct llist_head list;
 	struct sockaddr_in addr;
 	struct mgcp_config *cfg = ofd->data;
 	uint32_t rem;
-	struct mgcp_conn_rtp *conn_net = NULL;
+	struct mgcp_conn_rtp *conn_bts = NULL;
 
 	msg = osmux_recv(ofd, &addr);
 	if (!msg)
@@ -337,8 +344,8 @@ int osmux_read_from_bsc_nat_cb(struct osmo_fd *ofd, unsigned int what)
 				       &addr.sin_addr, MGCP_DEST_NET);
 
 		/* FIXME: Get rid of CONN_ID_XXX! */
-		conn_net = mgcp_conn_get_rtp(endp, CONN_ID_NET);
-		if (!conn_net)
+		conn_bts = mgcp_conn_get_rtp(endp, CONN_ID_BTS);
+		if (!conn_bts)
 			goto out;
 
 		if (!endp) {
@@ -347,12 +354,11 @@ int osmux_read_from_bsc_nat_cb(struct osmo_fd *ofd, unsigned int what)
 			     osmuxh->circuit_id);
 			goto out;
 		}
-		conn_net->osmux.stats.octets += osmux_chunk_length(msg, rem);
-		conn_net->osmux.stats.chunks++;
+		conn_bts->osmux.stats.octets += osmux_chunk_length(msg, rem);
+		conn_bts->osmux.stats.chunks++;
 		rem = msg->len;
 
-		osmux_xfrm_output(osmuxh, &conn_net->osmux.out, &list);
-		osmux_tx_sched(&list, scheduled_tx_bts_cb, endp);
+		osmux_xfrm_output_sched(&conn_bts->osmux.out, osmuxh);
 	}
 out:
 	msgb_free(msg);
@@ -418,7 +424,6 @@ int osmux_read_from_bsc_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	struct msgb *msg;
 	struct osmux_hdr *osmuxh;
-	struct llist_head list;
 	struct sockaddr_in addr;
 	struct mgcp_config *cfg = ofd->data;
 	uint32_t rem;
@@ -455,8 +460,7 @@ int osmux_read_from_bsc_cb(struct osmo_fd *ofd, unsigned int what)
 		conn_net->osmux.stats.chunks++;
 		rem = msg->len;
 
-		osmux_xfrm_output(osmuxh, &conn_net->osmux.out, &list);
-		osmux_tx_sched(&list, scheduled_tx_net_cb, endp);
+		osmux_xfrm_output_sched(&conn_net->osmux.out, osmuxh);
 	}
 out:
 	msgb_free(msg);
@@ -545,9 +549,13 @@ int osmux_enable_conn(struct mgcp_endpoint *endp, struct mgcp_conn_rtp *conn,
 	switch (endp->cfg->role) {
 		case MGCP_BSC_NAT:
 			conn->type = MGCP_OSMUX_BSC_NAT;
+			osmux_xfrm_output_set_tx_cb(&conn->osmux.out,
+							scheduled_tx_net_cb, endp);
 			break;
 		case MGCP_BSC:
 			conn->type = MGCP_OSMUX_BSC;
+			osmux_xfrm_output_set_tx_cb(&conn->osmux.out,
+							scheduled_tx_bts_cb, endp);
 			break;
 	}
 
@@ -566,8 +574,13 @@ void osmux_disable_conn(struct mgcp_conn_rtp *conn)
 	if (conn->osmux.state != OSMUX_STATE_ENABLED)
 		return;
 
-	LOGP(DLMGCP, LOGL_INFO, "Releasing connection %u using Osmux CID %u\n",
+	LOGP(DLMGCP, LOGL_INFO, "Releasing connection %s using Osmux CID %u\n",
 	     conn->conn->id, conn->osmux.cid);
+
+	/* We are closing, we don't need pending RTP packets to be transmitted */
+	osmux_xfrm_output_set_tx_cb(&conn->osmux.out, NULL, NULL);
+	osmux_xfrm_output_flush(&conn->osmux.out);
+
 	osmux_xfrm_input_close_circuit(conn->osmux.in, conn->osmux.cid);
 	conn->osmux.state = OSMUX_STATE_DISABLED;
 	conn->osmux.cid = -1;
