@@ -38,8 +38,9 @@
 #include <osmocom/mgcp/mgcp_internal.h>
 #include <osmocom/mgcp/mgcp_stat.h>
 #include <osmocom/mgcp/mgcp_msg.h>
-#include <osmocom/mgcp/mgcp_ep.h>
+#include <osmocom/mgcp/mgcp_endp.h>
 #include <osmocom/mgcp/mgcp_sdp.h>
+#include <osmocom/mgcp/mgcp_codec.h>
 
 struct mgcp_request {
 	char *name;
@@ -88,8 +89,7 @@ static int setup_rtp_processing(struct mgcp_endpoint *endp,
 		}
 	}
 
-	return cfg->setup_rtp_processing_cb(endp, &conn_dst->end,
-					    &conn_src->end);
+	return cfg->setup_rtp_processing_cb(endp, conn_dst, conn_src);
 }
 
 /* array of function pointers for handling various
@@ -192,17 +192,41 @@ static struct msgb *create_err_response(struct mgcp_endpoint *endp,
 	return create_resp(endp, code, " FAIL", msg, trans, NULL, NULL);
 }
 
+/* Add MGCP parameters to a message buffer */
+static int add_params(struct msgb *msg, const struct mgcp_endpoint *endp,
+		      const struct mgcp_conn_rtp *conn)
+{
+	int rc;
+
+	/* NOTE: Only in the virtual trunk we allow dynamic endpoint names */
+	if (endp->wildcarded_req
+	    && endp->tcfg->trunk_type == MGCP_TRUNK_VIRTUAL) {
+		rc = msgb_printf(msg, "Z: %s%x@%s\r\n",
+				 MGCP_ENDPOINT_PREFIX_VIRTUAL_TRUNK,
+				 ENDPOINT_NUMBER(endp), endp->cfg->domain);
+		if (rc < 0)
+			return -EINVAL;
+	}
+
+	rc = msgb_printf(msg, "I: %s\r\n", conn->conn->id);
+	if (rc < 0)
+		return -EINVAL;
+
+	return 0;
+}
+
 /* Format MGCP response string (with SDP attached) */
 static struct msgb *create_response_with_sdp(struct mgcp_endpoint *endp,
 					     struct mgcp_conn_rtp *conn,
 					     const char *msg,
-					     const char *trans_id)
+					     const char *trans_id,
+					     bool add_conn_params)
 {
 	const char *addr = endp->cfg->local_ip;
 	struct msgb *sdp;
 	int rc;
 	struct msgb *result;
-	char osmux_extension[strlen("\nX-Osmux: 255") + 1];
+	char osmux_extension[strlen("X-Osmux: 255") + 1];
 	char local_ip_addr[INET_ADDRSTRLEN];
 
 	sdp = msgb_alloc_headroom(4096, 128, "sdp record");
@@ -215,15 +239,28 @@ static struct msgb *create_response_with_sdp(struct mgcp_endpoint *endp,
 	}
 
 	if (conn->osmux.state == OSMUX_STATE_NEGOTIATING) {
-		sprintf(osmux_extension, "\nX-Osmux: %u", conn->osmux.cid);
+		sprintf(osmux_extension, "X-Osmux: %u", conn->osmux.cid);
 		conn->osmux.state = OSMUX_STATE_ACTIVATING;
 	} else {
 		osmux_extension[0] = '\0';
 	}
 
-	rc = msgb_printf(sdp, "I: %u%s\n\n", conn->conn->id, osmux_extension);
-	if (rc < 0)
-		goto error;
+	/* Attach optional connection parameters */
+	if (add_conn_params) {
+		rc = add_params(sdp, endp, conn);
+		if (rc < 0)
+			goto error;
+	}
+
+	/* Attach optional OSMUX parameters */
+	if (conn->osmux.state == OSMUX_STATE_NEGOTIATING) {
+		rc = msgb_printf(sdp, "%s\r\n", osmux_extension);
+		if (rc < 0)
+			goto error;
+	}
+
+	/* Attach line break to separate the parameters from the SDP block */
+	rc = msgb_printf(sdp, "\r\n");
 
 	rc = mgcp_write_response_sdp(endp, conn, sdp, addr);
 	if (rc < 0)
@@ -252,7 +289,7 @@ static void send_dummy(struct mgcp_endpoint *endp, struct mgcp_conn_rtp *conn)
 struct msgb *mgcp_handle_message(struct mgcp_config *cfg, struct msgb *msg)
 {
 	struct mgcp_parse_data pdata;
-	int i, code, handled = 0;
+	int rc, i, code, handled = 0;
 	struct msgb *resp = NULL;
 	char *data;
 
@@ -280,13 +317,19 @@ struct msgb *mgcp_handle_message(struct mgcp_config *cfg, struct msgb *msg)
 	memset(&pdata, 0, sizeof(pdata));
 	pdata.cfg = cfg;
 	data = mgcp_strline((char *)msg->l3h, &pdata.save);
-	pdata.found = mgcp_parse_header(&pdata, data);
+	rc = mgcp_parse_header(&pdata, data);
 	if (pdata.endp && pdata.trans
 	    && pdata.endp->last_trans
 	    && strcmp(pdata.endp->last_trans, pdata.trans) == 0) {
 		return do_retransmission(pdata.endp);
 	}
 
+	/* check for general parser failure */
+	if (rc < 0) {
+		LOGP(DLMGCP, LOGL_NOTICE, "%s: failed to find the endpoint\n", msg->l2h);
+		return create_err_response(NULL, -rc, (const char *) msg->l2h, pdata.trans);
+	}
+
 	for (i = 0; i < ARRAY_SIZE(mgcp_requests); ++i) {
 		if (strncmp
 		    (mgcp_requests[i].name, (const char *)&msg->l2h[0],
@@ -308,24 +351,20 @@ struct msgb *mgcp_handle_message(struct mgcp_config *cfg, struct msgb *msg)
 static struct msgb *handle_audit_endpoint(struct mgcp_parse_data *p)
 {
 	LOGP(DLMGCP, LOGL_NOTICE, "AUEP: auditing endpoint ...\n");
-
-	if (p->found != 0) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "AUEP: failed to find the endpoint.\n");
-		return create_err_response(NULL, 500, "AUEP", p->trans);
-	} else
-		return create_ok_response(p->endp, 200, "AUEP", p->trans);
+	return create_ok_response(p->endp, 200, "AUEP", p->trans);
 }
 
-/* Try to find a free port by attemting to bind on it. Also handle the
+/* Try to find a free port by attempting to bind on it. Also handle the
  * counter that points on the next free port. Since we have a pointer
- * to the next free port, binding should work on the first attemt,
- * neverless, try at least the next 200 ports before giving up */
+ * to the next free port, binding should in work on the first attempt in
+ * general. In case of failure the next port is tryed until the whole port
+ * range is tryed once. */
 static int allocate_port(struct mgcp_endpoint *endp, struct mgcp_conn_rtp *conn)
 {
 	int i;
 	struct mgcp_rtp_end *end;
 	struct mgcp_port_range *range;
+	unsigned int tries;
 
 	OSMO_ASSERT(conn);
 	end = &conn->end;
@@ -334,7 +373,8 @@ static int allocate_port(struct mgcp_endpoint *endp, struct mgcp_conn_rtp *conn)
 	range = &endp->cfg->net_ports;
 
 	/* attempt to find a port */
-	for (i = 0; i < 200; ++i) {
+	tries = (range->range_end - range->range_start) / 2;
+	for (i = 0; i < tries; ++i) {
 		int rc;
 
 		if (range->last_port >= range->range_end)
@@ -350,39 +390,177 @@ static int allocate_port(struct mgcp_endpoint *endp, struct mgcp_conn_rtp *conn)
 	}
 
 	LOGP(DLMGCP, LOGL_ERROR,
-	     "Allocating a RTP/RTCP port failed 200 times 0x%x.\n",
-	     ENDPOINT_NUMBER(endp));
+	     "Allocating a RTP/RTCP port failed %u times 0x%x.\n",
+	     tries, ENDPOINT_NUMBER(endp));
+	return -1;
+}
+
+/*! Helper function for check_local_cx_options() to get a pointer of the next
+ *  lco option identifier
+ *  \param[in] lco string
+ *  \returns pointer to the beginning of the LCO identifier, NULL on failure */
+char *get_lco_identifier(const char *options)
+{
+	char *ptr;
+	unsigned int count = 0;
+
+	/* Jump to the end of the lco identifier */
+	ptr = strstr(options, ":");
+	if (!ptr)
+		return NULL;
+
+	/* Walk backwards until the pointer points to the beginning of the
+	 * lco identifier. We know that we stand at the beginning when we
+	 * are either at the beginning of the memory or see a space or
+	 * comma. (this is tolerant, it will accept a:10, b:11 as well as
+	 * a:10,b:11) */
+	while (1) {
+		/* Endless loop protection */
+		if (count > 10000)
+			return NULL;
+		else if (ptr < options || *ptr == ' ' || *ptr == ',') {
+			ptr++;
+			break;
+		}
+		ptr--;
+		count++;
+	}
+
+	/* Check if we got any result */
+	if (*ptr == ':')
+		return NULL;
+
+	return ptr;
+}
+
+/*! Check the LCO option. This function checks for multiple appearence of LCO
+ *  options, which is illegal
+ *  \param[in] ctx talloc context
+ *  \param[in] lco string
+ *  \returns 0 on success, -1 on failure */
+int check_local_cx_options(void *ctx, const char *options)
+{
+	int i;
+	char *options_copy;
+	char *lco_identifier;
+	char *lco_identifier_end;
+	char *next_lco_identifier;
+
+	char **lco_seen;
+	unsigned int lco_seen_n = 0;
+
+	if (!options)
+		return -1;
+
+	lco_seen =
+	    (char **)talloc_zero_size(ctx, strlen(options) * sizeof(char *));
+	options_copy = talloc_strdup(ctx, options);
+	lco_identifier = options_copy;
+
+	do {
+		/* Move the lco_identifier pointer to the beginning of the
+		 * current lco option identifier */
+		lco_identifier = get_lco_identifier(lco_identifier);
+		if (!lco_identifier)
+			goto error;
+
+		/* Look ahead to the next LCO option early, since we
+		 * will parse destructively */
+		next_lco_identifier = strstr(lco_identifier + 1, ",");
+
+		/* Pinch off the end of the lco field identifier name
+		 * and see if we still got something, also check if
+		 * there is some value after the colon. */
+		lco_identifier_end = strstr(lco_identifier, ":");
+		if (!lco_identifier_end)
+			goto error;
+		if (*(lco_identifier_end + 1) == ' '
+		    || *(lco_identifier_end + 1) == ','
+		    || *(lco_identifier_end + 1) == '\0')
+			goto error;
+		*lco_identifier_end = '\0';
+		if (strlen(lco_identifier) == 0)
+			goto error;
+
+		/* Check if we have already seen the current field identifier
+		 * before. If yes, we must bail, an LCO must only appear once
+		 * in the LCO string */
+		for (i = 0; i < lco_seen_n; i++) {
+			if (strcmp(lco_seen[i], lco_identifier) == 0)
+				goto error;
+		}
+		lco_seen[lco_seen_n] = lco_identifier;
+		lco_seen_n++;
+
+		/* The first identifier must always be found at the beginnning
+		 * of the LCO string */
+		if (lco_seen[0] != options_copy)
+			goto error;
+
+		/* Go to the next lco option */
+		lco_identifier = next_lco_identifier;
+	} while (lco_identifier);
+
+	talloc_free(lco_seen);
+	talloc_free(options_copy);
+	return 0;
+error:
+	talloc_free(lco_seen);
+	talloc_free(options_copy);
 	return -1;
 }
 
 /* Set the LCO from a string (see RFC 3435).
- * The string is stored in the 'string' field. A NULL string is handled excatlyy
+ * The string is stored in the 'string' field. A NULL string is handled exactly
  * like an empty string, the 'string' field is never NULL after this function
  * has been called. */
-static void set_local_cx_options(void *ctx, struct mgcp_lco *lco,
+static int set_local_cx_options(void *ctx, struct mgcp_lco *lco,
 				 const char *options)
 {
 	char *p_opt, *a_opt;
 	char codec[9];
 
+	if (!options)
+		return 0;
+	if (strlen(options) == 0)
+		return 0;
+
+	/* Make sure the encoding of the LCO is consistant before we proceed */
+	if (check_local_cx_options(ctx, options) != 0) {
+		LOGP(DLMGCP, LOGL_ERROR,
+		     "local CX options: Internal inconsistency in Local Connection Options!\n");
+		return 524;
+	}
+
 	talloc_free(lco->string);
-	talloc_free(lco->codec);
-	lco->codec = NULL;
-	lco->pkt_period_min = lco->pkt_period_max = 0;
-	lco->string = talloc_strdup(ctx, options ? options : "");
+	lco->string = talloc_strdup(ctx, options);
 
 	p_opt = strstr(lco->string, "p:");
 	if (p_opt && sscanf(p_opt, "p:%d-%d",
 			    &lco->pkt_period_min, &lco->pkt_period_max) == 1)
 		lco->pkt_period_max = lco->pkt_period_min;
 
+	/* FIXME: LCO also supports the negotiation of more then one codec.
+	 * (e.g. a:PCMU;G726-32) But this implementation only supports a single
+	 * codec only. */
 	a_opt = strstr(lco->string, "a:");
-	if (a_opt && sscanf(a_opt, "a:%8[^,]", codec) == 1)
+	if (a_opt && sscanf(a_opt, "a:%8[^,]", codec) == 1) {
+		talloc_free(lco->codec);
 		lco->codec = talloc_strdup(ctx, codec);
+	}
 
 	LOGP(DLMGCP, LOGL_DEBUG,
 	     "local CX options: lco->pkt_period_max: %i, lco->codec: %s\n",
 	     lco->pkt_period_max, lco->codec);
+
+	/* Check if the packetization fits the 20ms raster */
+	if (lco->pkt_period_min % 20 && lco->pkt_period_max % 20) {
+		LOGP(DLMGCP, LOGL_ERROR,
+		     "local CX options: packetization interval is not a multiple of 20ms!\n");
+		return 535;
+	}
+
+	return 0;
 }
 
 void mgcp_rtp_end_config(struct mgcp_endpoint *endp, int expect_ssrc_change,
@@ -410,15 +588,15 @@ uint32_t mgcp_rtp_packet_duration(struct mgcp_endpoint *endp,
 	/* Get the number of frames per channel and packet */
 	if (rtp->frames_per_packet)
 		f = rtp->frames_per_packet;
-	else if (rtp->packet_duration_ms && rtp->codec.frame_duration_num) {
-		int den = 1000 * rtp->codec.frame_duration_num;
-		f = (rtp->packet_duration_ms * rtp->codec.frame_duration_den +
+	else if (rtp->packet_duration_ms && rtp->codec->frame_duration_num) {
+		int den = 1000 * rtp->codec->frame_duration_num;
+		f = (rtp->packet_duration_ms * rtp->codec->frame_duration_den +
 		     den / 2)
 		    / den;
 	}
 
-	return rtp->codec.rate * f * rtp->codec.frame_duration_num /
-	    rtp->codec.frame_duration_den;
+	return rtp->codec->rate * f * rtp->codec->frame_duration_num /
+	    rtp->codec->frame_duration_den;
 }
 
 static int mgcp_osmux_setup(struct mgcp_endpoint *endp, const char *line)
@@ -434,6 +612,68 @@ static int mgcp_osmux_setup(struct mgcp_endpoint *endp, const char *line)
 	return mgcp_parse_osmux_cid(line);
 }
 
+/* Process codec information contained in CRCX/MDCX */
+static int handle_codec_info(struct mgcp_conn_rtp *conn,
+			     struct mgcp_parse_data *p, int have_sdp, bool crcx)
+{
+	struct mgcp_endpoint *endp = p->endp;
+	int rc;
+	char *cmd;
+
+	if (crcx)
+		cmd = "CRCX";
+	else
+		cmd = "MDCX";
+
+	/* Collect codec information */
+	if (have_sdp) {
+		/* If we have SDP, we ignore the local connection options and
+		 * use only the SDP information. */
+		mgcp_codec_reset_all(conn);
+		rc = mgcp_parse_sdp_data(endp, conn, p);
+		if (rc != 0) {
+			LOGP(DLMGCP, LOGL_ERROR,
+			     "%s: endpoint:%x sdp not parseable\n", cmd,
+			     ENDPOINT_NUMBER(endp));
+
+			/* See also RFC 3661: Protocol error */
+			return 510;
+		}
+	} else if (endp->local_options.codec) {
+		/* When no SDP is available, we use the codec information from
+		 * the local connection options (if present) */
+		mgcp_codec_reset_all(conn);
+		rc = mgcp_codec_add(conn, PTYPE_UNDEFINED, endp->local_options.codec);
+		if (rc != 0)
+			goto error;
+	}
+
+	/* Make sure we always set a sane default codec */
+	if (conn->end.codecs_assigned == 0) {
+		/* When SDP and/or LCO did not supply any codec information,
+		 * than it makes sense to pick a sane default: (payload-type 0,
+		 * PCMU), see also: OS#2658 */
+		mgcp_codec_reset_all(conn);
+		rc = mgcp_codec_add(conn, 0, NULL);
+		if (rc != 0)
+			goto error;
+	}
+
+	/* Make codec decision */
+	if (mgcp_codec_decide(conn) != 0)
+		goto error;
+
+	return 0;
+
+error:
+	LOGP(DLMGCP, LOGL_ERROR,
+	     "%s: endpoint:0x%x codec negotiation failure\n", cmd,
+	     ENDPOINT_NUMBER(endp));
+
+	/* See also RFC 3661: Codec negotiation failure */
+	return 534;
+}
+
 /* CRCX command handler, processes the received command */
 static struct msgb *handle_create_con(struct mgcp_parse_data *p)
 {
@@ -443,19 +683,16 @@ static struct msgb *handle_create_con(struct mgcp_parse_data *p)
 
 	const char *local_options = NULL;
 	const char *callid = NULL;
-	const char *ci = NULL;
 	const char *mode = NULL;
 	char *line;
 	int have_sdp = 0, osmux_cid = -1;
 	struct mgcp_conn_rtp *conn = NULL;
-	uint32_t conn_id;
+	struct mgcp_conn *_conn = NULL;
 	char conn_name[512];
+	int rc;
 
 	LOGP(DLMGCP, LOGL_NOTICE, "CRCX: creating new connection ...\n");
 
-	if (p->found != 0)
-		return create_err_response(NULL, 510, "CRCX", p->trans);
-
 	/* parse CallID C: and LocalParameters L: */
 	for_each_line(line, p->save) {
 		if (!mgcp_check_param(endp, line))
@@ -469,7 +706,10 @@ static struct msgb *handle_create_con(struct mgcp_parse_data *p)
 			callid = (const char *)line + 3;
 			break;
 		case 'I':
-			ci = (const char *)line + 3;
+			/* It is illegal to send a connection identifier
+			 * together with a CRCX, the MGW will assign the
+			 * connection identifier by itself on CRCX */
+			return create_err_response(NULL, 523, "CRCX", p->trans);
 			break;
 		case 'M':
 			mode = (const char *)line + 3;
@@ -489,6 +729,7 @@ static struct msgb *handle_create_con(struct mgcp_parse_data *p)
 			LOGP(DLMGCP, LOGL_NOTICE,
 			     "CRCX: endpoint:%x unhandled option: '%c'/%d\n",
 			     ENDPOINT_NUMBER(endp), *line, *line);
+			return create_err_response(NULL, 539, "CRCX", p->trans);
 			break;
 		}
 	}
@@ -501,21 +742,14 @@ mgcp_header_done:
 		LOGP(DLMGCP, LOGL_ERROR,
 		     "CRCX: endpoint:%x insufficient parameters, missing callid\n",
 		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "CRCX", p->trans);
+		return create_err_response(endp, 516, "CRCX", p->trans);
 	}
 
 	if (!mode) {
 		LOGP(DLMGCP, LOGL_ERROR,
 		     "CRCX: endpoint:%x insufficient parameters, missing mode\n",
 		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "CRCX", p->trans);
-	}
-
-	if (!ci) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x insufficient parameters, missing connection id\n",
-		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "CRCX", p->trans);
+		return create_err_response(endp, 517, "CRCX", p->trans);
 	}
 
 	/* Check if we are able to accept the creation of another connection */
@@ -531,7 +765,7 @@ mgcp_header_done:
 		} else {
 			/* There is no more room for a connection, leave
 			 * everything as it is and return with an error */
-			return create_err_response(endp, 400, "CRCX", p->trans);
+			return create_err_response(endp, 540, "CRCX", p->trans);
 		}
 	}
 
@@ -539,12 +773,12 @@ mgcp_header_done:
 	 * callids match up so that we are sure that this is our call */
 	if (endp->callid && mgcp_verify_call_id(endp, callid)) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x allready seized by other call (%s)\n",
+		     "CRCX: endpoint:0x%x allready seized by other call (%s)\n",
 		     ENDPOINT_NUMBER(endp), endp->callid);
 		if (tcfg->force_realloc)
 			/* This is not our call, toss everything by releasing
 			 * the entire endpoint. (rude!) */
-			mgcp_release_endp(endp);
+			mgcp_endp_release(endp);
 		else {
 			/* This is not our call, leave everything as it is and
 			 * return with an error. */
@@ -553,47 +787,21 @@ mgcp_header_done:
 	}
 
 	/* Set the callid, creation of another connection will only be possible
-	 * when the callid matches up. (Connections are distinuished by their
+	 * when the callid matches up. (Connections are distinguished by their
 	 * connection ids) */
 	endp->callid = talloc_strdup(tcfg->endpoints, callid);
 
-	/* Extract audio codec information */
-	set_local_cx_options(endp->tcfg->endpoints, &endp->local_options,
-			     local_options);
-
-	if (mgcp_parse_ci(&conn_id, ci)) {
+	snprintf(conn_name, sizeof(conn_name), "%s", callid);
+	_conn = mgcp_conn_alloc(NULL, endp, MGCP_CONN_TYPE_RTP, conn_name);
+	if (!_conn) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x insufficient parameters, missing ci (connectionIdentifier)\n",
-		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "CRCX", p->trans);
-	}
-
-	/* Only accept another connection when the connection ID is different. */
-	if (mgcp_conn_get_rtp(endp, conn_id)) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x there is already a connection with id %u present!\n",
-		     conn_id, ENDPOINT_NUMBER(endp));
-		if (tcfg->force_realloc) {
-			/* Ignore the existing connection by just freeing it */
-			mgcp_conn_free(endp, conn_id);
-		} else {
-			/* There is already a connection with that ID present,
-			 * leave everything as it is and return with an error. */
-			return create_err_response(endp, 400, "CRCX", p->trans);
-		}
-	}
-
-	snprintf(conn_name, sizeof(conn_name), "%s-%u", callid, conn_id);
-	mgcp_conn_alloc(NULL, endp, conn_id, MGCP_CONN_TYPE_RTP,
-			conn_name);
-	conn = mgcp_conn_get_rtp(endp, conn_id);
-	if (!conn) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x unable to allocate RTP connection\n",
+		     "CRCX: endpoint:0x%x unable to allocate RTP connection\n",
 		     ENDPOINT_NUMBER(endp));
 		goto error2;
 
 	}
+	conn = mgcp_conn_get_rtp(endp, _conn->id);
+	OSMO_ASSERT(conn);
 
 	if (mgcp_parse_conn_mode(mode, endp, conn->conn) != 0) {
 		error_code = 517;
@@ -608,17 +816,32 @@ mgcp_header_done:
 		conn->osmux.state = OSMUX_STATE_NEGOTIATING;
 	} else if (endp->cfg->osmux == OSMUX_USAGE_ONLY) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x osmux only and no osmux offered\n",
+		     "CRCX: endpoint:0x%x osmux only and no osmux offered\n",
 		     ENDPOINT_NUMBER(endp));
 		goto error2;
 	}
 
-	/* set up RTP media parameters */
-	if (have_sdp)
-		mgcp_parse_sdp_data(endp, conn, p);
-	else if (endp->local_options.codec)
-		mgcp_set_audio_info(p->cfg, &conn->end.codec,
-				    PTYPE_UNDEFINED, endp->local_options.codec);
+	/* Set local connection options, if present */
+	if (local_options) {
+		rc = set_local_cx_options(endp->tcfg->endpoints,
+					  &endp->local_options, local_options);
+		if (rc != 0) {
+			LOGP(DLMGCP, LOGL_ERROR,
+			     "CRCX: endpoint:%x inavlid local connection options!\n",
+			     ENDPOINT_NUMBER(endp));
+			error_code = rc;
+			goto error2;
+		}
+	}
+
+	/* Handle codec information and decide for a suitable codec */
+	rc = handle_codec_info(conn, p, have_sdp, true);
+	mgcp_codec_summary(conn);
+	if (rc) {
+		error_code = rc;
+		goto error2;
+	}
+
 	conn->end.fmtp_extra = talloc_strdup(tcfg->endpoints,
 					     tcfg->audio_fmtp_extra);
 
@@ -629,13 +852,24 @@ mgcp_header_done:
 
 	mgcp_rtp_end_config(endp, 0, &conn->end);
 
+	/* check connection mode setting */
+	if (conn->conn->mode != MGCP_CONN_LOOPBACK
+	    && conn->conn->mode != MGCP_CONN_RECV_ONLY
+	    && conn->end.rtp_port == 0) {
+		LOGP(DLMGCP, LOGL_ERROR,
+		     "CRCX: endpoint:%x selected connection mode type requires an opposite end!\n",
+		     ENDPOINT_NUMBER(endp));
+		error_code = 527;
+		goto error2;
+	}
+
 	if (allocate_port(endp, conn) != 0) {
 		goto error2;
 	}
 
 	if (setup_rtp_processing(endp, conn) != 0) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "CRCX: endpoint:%x could not start RTP processing!\n",
+		     "CRCX: endpoint:0x%x could not start RTP processing!\n",
 		     ENDPOINT_NUMBER(endp));
 		goto error2;
 	}
@@ -648,9 +882,9 @@ mgcp_header_done:
 		switch (rc) {
 		case MGCP_POLICY_REJECT:
 			LOGP(DLMGCP, LOGL_NOTICE,
-			     "CRCX: endpoint:%x CRCX rejected by policy\n",
+			     "CRCX: endpoint:0x%x CRCX rejected by policy\n",
 			     ENDPOINT_NUMBER(endp));
-			mgcp_release_endp(endp);
+			mgcp_endp_release(endp);
 			return create_err_response(endp, 400, "CRCX", p->trans);
 			break;
 		case MGCP_POLICY_DEFER:
@@ -664,7 +898,7 @@ mgcp_header_done:
 	}
 
 	LOGP(DLMGCP, LOGL_DEBUG,
-	     "CRCX: endpoint:%x Creating connection: CI: %u port: %u\n",
+	     "CRCX: endpoint:0x%x Creating connection: CI: %s port: %u\n",
 	     ENDPOINT_NUMBER(endp), conn->conn->id, conn->end.local_port);
 	if (p->cfg->change_cb)
 		p->cfg->change_cb(tcfg, ENDPOINT_NUMBER(endp), MGCP_ENDP_CRCX);
@@ -676,17 +910,21 @@ mgcp_header_done:
 		send_dummy(endp, conn);
 
 	LOGP(DLMGCP, LOGL_NOTICE,
-	     "CRCX: endpoint:%x connection successfully created\n",
+	     "CRCX: endpoint:0x%x connection successfully created\n",
 	     ENDPOINT_NUMBER(endp));
-	return create_response_with_sdp(endp, conn, "CRCX", p->trans);
+	return create_response_with_sdp(endp, conn, "CRCX", p->trans, true);
 error2:
-	mgcp_release_endp(endp);
+	mgcp_endp_release(endp);
 	LOGP(DLMGCP, LOGL_NOTICE,
-	     "CRCX: endpoint:%x unable to create connection resource error\n",
+	     "CRCX: endpoint:0x%x unable to create connection\n",
 	     ENDPOINT_NUMBER(endp));
 	return create_err_response(endp, error_code, "CRCX", p->trans);
 }
 
+
+
+
+
 /* MDCX command handler, processes the received command */
 static struct msgb *handle_modify_con(struct mgcp_parse_data *p)
 {
@@ -695,20 +933,25 @@ static struct msgb *handle_modify_con(struct mgcp_parse_data *p)
 	int silent = 0;
 	int have_sdp = 0;
 	char *line;
-	const char *ci = NULL;
 	const char *local_options = NULL;
 	const char *mode = NULL;
 	struct mgcp_conn_rtp *conn = NULL;
-	uint32_t conn_id;
+        const char *conn_id = NULL;
+	int rc;
 
 	LOGP(DLMGCP, LOGL_NOTICE, "MDCX: modifying existing connection ...\n");
 
-	if (p->found != 0)
-		return create_err_response(NULL, 510, "MDCX", p->trans);
+	/* Prohibit wildcarded requests */
+	if (endp->wildcarded_req) {
+		LOGP(DLMGCP, LOGL_ERROR,
+		     "MDCX: endpoint:0x%x wildcarded endpoint names not supported.\n",
+		     ENDPOINT_NUMBER(endp));
+		return create_err_response(endp, 507, "MDCX", p->trans);
+	}
 
 	if (llist_count(&endp->conns) <= 0) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "MDCX: endpoint:%x endpoint is not holding a connection.\n",
+		     "MDCX: endpoint:0x%x endpoint is not holding a connection.\n",
 		     ENDPOINT_NUMBER(endp));
 		return create_err_response(endp, 400, "MDCX", p->trans);
 	}
@@ -719,13 +962,17 @@ static struct msgb *handle_modify_con(struct mgcp_parse_data *p)
 
 		switch (line[0]) {
 		case 'C':
-			if (mgcp_verify_call_id(endp, line + 3) != 0)
+			if (mgcp_verify_call_id(endp, line + 3) != 0) {
+				error_code = 516;
 				goto error3;
+			}
 			break;
 		case 'I':
-			ci = (const char *)line + 3;
-			if (mgcp_verify_ci(endp, ci) != 0)
+			conn_id = (const char *)line + 3;
+			if (mgcp_verify_ci(endp, conn_id) != 0) {
+				error_code = 515;
 				goto error3;
+			}
 			break;
 		case 'L':
 			local_options = (const char *)line + 3;
@@ -742,18 +989,19 @@ static struct msgb *handle_modify_con(struct mgcp_parse_data *p)
 			break;
 		default:
 			LOGP(DLMGCP, LOGL_NOTICE,
-			     "MDCX: endpoint:%x Unhandled MGCP option: '%c'/%d\n",
+			     "MDCX: endpoint:0x%x Unhandled MGCP option: '%c'/%d\n",
 			     ENDPOINT_NUMBER(endp), line[0], line[0]);
+			return create_err_response(NULL, 539, "MDCX", p->trans);
 			break;
 		}
 	}
 
 mgcp_header_done:
-	if (mgcp_parse_ci(&conn_id, ci)) {
+	if (!conn_id) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "MDCX: endpoint:%x insufficient parameters, missing ci (connectionIdentifier)\n",
+		     "MDCX: endpoint:0x%x insufficient parameters, missing ci (connectionIdentifier)\n",
 		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "MDCX", p->trans);
+		return create_err_response(endp, 515, "MDCX", p->trans);
 	}
 
 	conn = mgcp_conn_get_rtp(endp, conn_id);
@@ -768,15 +1016,38 @@ mgcp_header_done:
 	} else
 			conn->conn->mode = conn->conn->mode_orig;
 
-	if (have_sdp)
-		mgcp_parse_sdp_data(endp, conn, p);
+	/* Set local connection options, if present */
+	if (local_options) {
+		rc = set_local_cx_options(endp->tcfg->endpoints,
+					  &endp->local_options, local_options);
+		if (rc != 0) {
+			LOGP(DLMGCP, LOGL_ERROR,
+			     "MDCX: endpoint:%x inavlid local connection options!\n",
+			     ENDPOINT_NUMBER(endp));
+			error_code = rc;
+			goto error3;
+		}
+	}
 
-	set_local_cx_options(endp->tcfg->endpoints, &endp->local_options,
-			     local_options);
+	/* Handle codec information and decide for a suitable codec */
+	rc = handle_codec_info(conn, p, have_sdp, false);
+	mgcp_codec_summary(conn);
+	if (rc) {
+		error_code = rc;
+		goto error3;
+	}
+
+	/* check connection mode setting */
+	if (conn->conn->mode != MGCP_CONN_LOOPBACK
+	    && conn->conn->mode != MGCP_CONN_RECV_ONLY
+	    && conn->end.rtp_port == 0) {
+		LOGP(DLMGCP, LOGL_ERROR,
+		     "MDCX: endpoint:%x selected connection mode type requires an opposite end!\n",
+		     ENDPOINT_NUMBER(endp));
+		error_code = 527;
+		goto error3;
+	}
 
-	if (!have_sdp && endp->local_options.codec)
-		mgcp_set_audio_info(p->cfg, &conn->end.codec,
-				    PTYPE_UNDEFINED, endp->local_options.codec);
 
 	if (setup_rtp_processing(endp, conn) != 0)
 		goto error3;
@@ -790,7 +1061,7 @@ mgcp_header_done:
 		switch (rc) {
 		case MGCP_POLICY_REJECT:
 			LOGP(DLMGCP, LOGL_NOTICE,
-			     "MDCX: endpoint:%x rejected by policy\n",
+			     "MDCX: endpoint:0x%x rejected by policy\n",
 			     ENDPOINT_NUMBER(endp));
 			if (silent)
 				goto out_silent;
@@ -799,7 +1070,7 @@ mgcp_header_done:
 		case MGCP_POLICY_DEFER:
 			/* stop processing */
 			LOGP(DLMGCP, LOGL_DEBUG,
-			     "MDCX: endpoint:%x defered by policy\n",
+			     "MDCX: endpoint:0x%x defered by policy\n",
 			     ENDPOINT_NUMBER(endp));
 			return NULL;
 			break;
@@ -813,7 +1084,7 @@ mgcp_header_done:
 
 	/* modify */
 	LOGP(DLMGCP, LOGL_DEBUG,
-	     "MDCX: endpoint:%x modified conn:%s\n",
+	     "MDCX: endpoint:0x%x modified conn:%s\n",
 	     ENDPOINT_NUMBER(endp), mgcp_conn_dump(conn->conn));
 	if (p->cfg->change_cb)
 		p->cfg->change_cb(endp->tcfg, ENDPOINT_NUMBER(endp),
@@ -829,14 +1100,14 @@ mgcp_header_done:
 		goto out_silent;
 
 	LOGP(DLMGCP, LOGL_NOTICE,
-	     "MDCX: endpoint:%x connection successfully modified\n",
+	     "MDCX: endpoint:0x%x connection successfully modified\n",
 	     ENDPOINT_NUMBER(endp));
-	return create_response_with_sdp(endp, conn, "MDCX", p->trans);
+	return create_response_with_sdp(endp, conn, "MDCX", p->trans, false);
 error3:
 	return create_err_response(endp, error_code, "MDCX", p->trans);
 
 out_silent:
-	LOGP(DLMGCP, LOGL_DEBUG, "MDCX: endpoint:%x silent exit\n",
+	LOGP(DLMGCP, LOGL_DEBUG, "MDCX: endpoint:0x%x silent exit\n",
 	     ENDPOINT_NUMBER(endp));
 	return NULL;
 }
@@ -849,22 +1120,26 @@ static struct msgb *handle_delete_con(struct mgcp_parse_data *p)
 	int silent = 0;
 	char *line;
 	char stats[1048];
-	const char *ci = NULL;
+	const char *conn_id = NULL;
 	struct mgcp_conn_rtp *conn = NULL;
-	uint32_t conn_id;
 
 	LOGP(DLMGCP, LOGL_NOTICE,
-	     "DLCX: endpoint:%x deleting connection ...\n",
+	     "DLCX: endpoint:0x%x deleting connection ...\n",
 	     ENDPOINT_NUMBER(endp));
 
-	if (p->found != 0)
-		return create_err_response(NULL, error_code, "DLCX", p->trans);
+	/* Prohibit wildcarded requests */
+	if (endp->wildcarded_req) {
+		LOGP(DLMGCP, LOGL_ERROR,
+		     "DLCX: endpoint:0x%x wildcarded endpoint names not supported.\n",
+		     ENDPOINT_NUMBER(endp));
+		return create_err_response(endp, 507, "DLCX", p->trans);
+	}
 
 	if (llist_count(&endp->conns) <= 0) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "DLCX: endpoint:%x endpoint is not holding a connection.\n",
+		     "DLCX: endpoint:0x%x endpoint is not holding a connection.\n",
 		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "DLCX", p->trans);
+		return create_err_response(endp, 515, "DLCX", p->trans);
 	}
 
 	for_each_line(line, p->save) {
@@ -873,21 +1148,26 @@ static struct msgb *handle_delete_con(struct mgcp_parse_data *p)
 
 		switch (line[0]) {
 		case 'C':
-			if (mgcp_verify_call_id(endp, line + 3) != 0)
+			if (mgcp_verify_call_id(endp, line + 3) != 0) {
+				error_code = 516;
 				goto error3;
+			}
 			break;
 		case 'I':
-			ci = (const char *)line + 3;
-			if (mgcp_verify_ci(endp, ci) != 0)
+			conn_id = (const char *)line + 3;
+			if (mgcp_verify_ci(endp, conn_id) != 0) {
+				error_code = 515;
 				goto error3;
+			}
 			break;
 		case 'Z':
 			silent = strcmp("noanswer", line + 3) == 0;
 			break;
 		default:
 			LOGP(DLMGCP, LOGL_NOTICE,
-			     "DLCX: endpoint:%x Unhandled MGCP option: '%c'/%d\n",
+			     "DLCX: endpoint:0x%x Unhandled MGCP option: '%c'/%d\n",
 			     ENDPOINT_NUMBER(endp), line[0], line[0]);
+			return create_err_response(NULL, 539, "DLCX", p->trans);
 			break;
 		}
 	}
@@ -900,7 +1180,7 @@ static struct msgb *handle_delete_con(struct mgcp_parse_data *p)
 		switch (rc) {
 		case MGCP_POLICY_REJECT:
 			LOGP(DLMGCP, LOGL_NOTICE,
-			     "DLCX: endpoint:%x rejected by policy\n",
+			     "DLCX: endpoint:0x%x rejected by policy\n",
 			     ENDPOINT_NUMBER(endp));
 			if (silent)
 				goto out_silent;
@@ -919,12 +1199,12 @@ static struct msgb *handle_delete_con(struct mgcp_parse_data *p)
 	/* When no connection id is supplied, we will interpret this as a
 	 * wildcarded DLCX and drop all connections at once. (See also
 	 * RFC3435 Section F.7) */
-	if (!ci) {
+	if (!conn_id) {
 		LOGP(DLMGCP, LOGL_NOTICE,
-		     "DLCX: endpoint:%x missing ci (connectionIdentifier), will remove all connections at once\n",
+		     "DLCX: endpoint:0x%x missing ci (connectionIdentifier), will remove all connections at once\n",
 		     ENDPOINT_NUMBER(endp));
 
-		mgcp_release_endp(endp);
+		mgcp_endp_release(endp);
 
 		/* Note: In this case we do not return any statistics,
 		 * as we assume that the client is not interested in
@@ -932,14 +1212,6 @@ static struct msgb *handle_delete_con(struct mgcp_parse_data *p)
 		return create_ok_response(endp, 200, "DLCX", p->trans);
 	}
 
-	/* Parse the connection id */
-	if (mgcp_parse_ci(&conn_id, ci)) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "DLCX: endpoint:%x insufficient parameters, invalid ci (connectionIdentifier)\n",
-		     ENDPOINT_NUMBER(endp));
-		return create_err_response(endp, 400, "DLCX", p->trans);
-	}
-
 	/* Find the connection */
 	conn = mgcp_conn_get_rtp(endp, conn_id);
 	if (!conn)
@@ -949,19 +1221,19 @@ static struct msgb *handle_delete_con(struct mgcp_parse_data *p)
 	mgcp_format_stats(stats, sizeof(stats), conn->conn);
 
 	/* delete connection */
-	LOGP(DLMGCP, LOGL_DEBUG, "DLCX: endpoint:%x deleting conn:%s\n",
+	LOGP(DLMGCP, LOGL_DEBUG, "DLCX: endpoint:0x%x deleting conn:%s\n",
 	     ENDPOINT_NUMBER(endp), mgcp_conn_dump(conn->conn));
 	mgcp_conn_free(endp, conn_id);
 	LOGP(DLMGCP, LOGL_NOTICE,
-	     "DLCX: endpoint:%x connection successfully deleted\n",
+	     "DLCX: endpoint:0x%x connection successfully deleted\n",
 	     ENDPOINT_NUMBER(endp));
 
 	/* When all connections are closed, the endpoint will be released
 	 * in order to be ready to be used by another call. */
 	if (llist_count(&endp->conns) <= 0) {
-		mgcp_release_endp(endp);
+		mgcp_endp_release(endp);
 		LOGP(DLMGCP, LOGL_DEBUG,
-		     "DLCX: endpoint:%x endpoint released\n",
+		     "DLCX: endpoint:0x%x endpoint released\n",
 		     ENDPOINT_NUMBER(endp));
 	}
 
@@ -977,7 +1249,7 @@ error3:
 	return create_err_response(endp, error_code, "DLCX", p->trans);
 
 out_silent:
-	LOGP(DLMGCP, LOGL_DEBUG, "DLCX: endpoint:%x silent exit\n",
+	LOGP(DLMGCP, LOGL_DEBUG, "DLCX: endpoint:0x%x silent exit\n",
 	     ENDPOINT_NUMBER(endp));
 	return NULL;
 }
@@ -995,12 +1267,6 @@ static struct msgb *handle_rsip(struct mgcp_parse_data *p)
 
 	LOGP(DLMGCP, LOGL_NOTICE, "RSIP: resetting all endpoints ...\n");
 
-	if (p->found != 0) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "RSIP: failed to find the endpoint.\n");
-		return NULL;
-	}
-
 	if (p->cfg->reset_cb)
 		p->cfg->reset_cb(p->endp->tcfg);
 	return NULL;
@@ -1026,9 +1292,6 @@ static struct msgb *handle_noti_req(struct mgcp_parse_data *p)
 
 	LOGP(DLMGCP, LOGL_NOTICE, "RQNT: processing request for notification ...\n");
 
-	if (p->found != 0)
-		return create_err_response(NULL, 400, "RQNT", p->trans);
-
 	for_each_line(line, p->save) {
 		switch (line[0]) {
 		case 'S':
@@ -1050,7 +1313,7 @@ static struct msgb *handle_noti_req(struct mgcp_parse_data *p)
 }
 
 /* Connection keepalive timer, will take care that dummy packets are send
- * regulary, so that NAT connections stay open */
+ * regularly, so that NAT connections stay open */
 static void mgcp_keepalive_timer_cb(void *_tcfg)
 {
 	struct mgcp_trunk_config *tcfg = _tcfg;
@@ -1119,6 +1382,8 @@ struct mgcp_config *mgcp_config_alloc(void)
 		return NULL;
 	}
 
+	osmo_strlcpy(cfg->domain, "mgw", sizeof(cfg->domain));
+
 	cfg->net_ports.range_start = RTP_PORT_DEFAULT_RANGE_START;
 	cfg->net_ports.range_end = RTP_PORT_DEFAULT_RANGE_END;
 	cfg->net_ports.last_port = cfg->net_ports.range_start;
@@ -1170,7 +1435,7 @@ struct mgcp_trunk_config *mgcp_trunk_alloc(struct mgcp_config *cfg, int nr)
 	trunk->audio_payload = 126;
 	trunk->audio_send_ptime = 1;
 	trunk->audio_send_name = 1;
-	trunk->number_endpoints = 33;
+	trunk->vty_number_endpoints = 33;
 	trunk->omit_rtcp = 0;
 	mgcp_trunk_set_keepalive(trunk, MGCP_KEEPALIVE_ONCE);
 	llist_add_tail(&trunk->entry, &cfg->trunks);
@@ -1202,12 +1467,12 @@ int mgcp_endpoints_allocate(struct mgcp_trunk_config *tcfg)
 
 	tcfg->endpoints = _talloc_zero_array(tcfg->cfg,
 					     sizeof(struct mgcp_endpoint),
-					     tcfg->number_endpoints,
+					     tcfg->vty_number_endpoints,
 					     "endpoints");
 	if (!tcfg->endpoints)
 		return -1;
 
-	for (i = 0; i < tcfg->number_endpoints; ++i) {
+	for (i = 0; i < tcfg->vty_number_endpoints; ++i) {
 		INIT_LLIST_HEAD(&tcfg->endpoints[i].conns);
 		tcfg->endpoints[i].cfg = tcfg->cfg;
 		tcfg->endpoints[i].tcfg = tcfg;
@@ -1217,31 +1482,10 @@ int mgcp_endpoints_allocate(struct mgcp_trunk_config *tcfg)
 		tcfg->endpoints[i].type = &ep_typeset.rtp;
 	}
 
+	tcfg->number_endpoints = tcfg->vty_number_endpoints;
 	return 0;
 }
 
-/*! relase endpoint, all open connections are closed.
- *  \param[in] endp endpoint to release */
-void mgcp_release_endp(struct mgcp_endpoint *endp)
-{
-	LOGP(DLMGCP, LOGL_DEBUG, "Releasing endpoint:%x\n",
-	     ENDPOINT_NUMBER(endp));
-
-	/* Normally this function should only be called wehen
-	 * all connections have been removed already. In case
-	 * that there are still connections open (e.g. when
-	 * RSIP is executed), free them all at once. */
-	mgcp_conn_free_all(endp);
-
-	/* Reset endpoint parameters and states */
-	talloc_free(endp->callid);
-	endp->callid = NULL;
-	talloc_free(endp->local_options.string);
-	endp->local_options.string = NULL;
-	talloc_free(endp->local_options.codec);
-	endp->local_options.codec = NULL;
-}
-
 static int send_agent(struct mgcp_config *cfg, const char *buf, int len)
 {
 	return write(cfg->gw_fd.bfd.fd, buf, len);
@@ -1254,13 +1498,16 @@ static int send_agent(struct mgcp_config *cfg, const char *buf, int len)
  *  \returns 0 on success, -1 on error */
 int mgcp_send_reset_all(struct mgcp_config *cfg)
 {
+	char buf[MGCP_ENDPOINT_MAXLEN + 128];
+	int len;
 	int rc;
 
-	static const char mgcp_reset[] = {
-		"RSIP 1 *@mgw MGCP 1.0\r\n"
-	};
+	len = snprintf(buf, sizeof(buf),
+		       "RSIP 1 *@%s MGCP 1.0\r\n", cfg->domain);
+	if (len < 0)
+		return -1;
 
-	rc = send_agent(cfg, mgcp_reset, sizeof mgcp_reset - 1);
+	rc = send_agent(cfg, buf, len);
 	if (rc <= 0)
 		return -1;
 
@@ -1274,17 +1521,15 @@ int mgcp_send_reset_all(struct mgcp_config *cfg)
  *  \returns 0 on success, -1 on error */
 int mgcp_send_reset_ep(struct mgcp_endpoint *endp, int endpoint)
 {
-	char buf[128];
+	char buf[MGCP_ENDPOINT_MAXLEN + 128];
 	int len;
 	int rc;
 
 	len = snprintf(buf, sizeof(buf),
-		       "RSIP 39 %x@mgw MGCP 1.0\r\n", endpoint);
+		       "RSIP 39 %x@%s MGCP 1.0\r\n", endpoint, endp->cfg->domain);
 	if (len < 0)
 		return -1;
 
-	buf[sizeof(buf) - 1] = '\0';
-
 	rc = send_agent(endp->cfg, buf, len);
 	if (rc <= 0)
 		return -1;
