@@ -27,8 +27,10 @@
 #include <osmocom/mgcp/mgcp_internal.h>
 #include <osmocom/mgcp/vty.h>
 #include <osmocom/mgcp/mgcp_conn.h>
+#include <osmocom/mgcp/mgcp_endp.h>
 
 #include <string.h>
+#include <inttypes.h>
 
 #define RTCP_OMIT_STR "Drop RTCP packets in both directions\n"
 #define RTP_PATCH_STR "Modify RTP packet header in both directions\n"
@@ -63,15 +65,16 @@ struct cmd_node trunk_node = {
 static int config_write_mgcp(struct vty *vty)
 {
 	vty_out(vty, "mgcp%s", VTY_NEWLINE);
+	vty_out(vty, "  domain %s%s", g_cfg->domain, VTY_NEWLINE);
 	if (g_cfg->local_ip)
 		vty_out(vty, "  local ip %s%s", g_cfg->local_ip, VTY_NEWLINE);
 	vty_out(vty, "  bind ip %s%s", g_cfg->source_addr, VTY_NEWLINE);
 	vty_out(vty, "  bind port %u%s", g_cfg->source_port, VTY_NEWLINE);
-	vty_out(vty, "  rtp net-range %u %u%s",
+	vty_out(vty, "  rtp port-range %u %u%s",
 		g_cfg->net_ports.range_start, g_cfg->net_ports.range_end,
 		VTY_NEWLINE);
 	if (g_cfg->net_ports.bind_addr)
-		vty_out(vty, "  rtp net-bind-ip %s%s",
+		vty_out(vty, "  rtp bind-ip %s%s",
 			g_cfg->net_ports.bind_addr, VTY_NEWLINE);
 	if (g_cfg->net_ports.bind_addr_probe)
 		vty_out(vty, "  rtp ip-probing%s", VTY_NEWLINE);
@@ -115,7 +118,7 @@ static int config_write_mgcp(struct vty *vty)
 		g_cfg->trunk.audio_send_name ? "" : "no ", VTY_NEWLINE);
 	vty_out(vty, "  loop %u%s", ! !g_cfg->trunk.audio_loop, VTY_NEWLINE);
 	vty_out(vty, "  number endpoints %u%s",
-		g_cfg->trunk.number_endpoints - 1, VTY_NEWLINE);
+		g_cfg->trunk.vty_number_endpoints - 1, VTY_NEWLINE);
 	vty_out(vty, "  %sallow-transcoding%s",
 		g_cfg->trunk.no_audio_transcoding ? "no " : "", VTY_NEWLINE);
 	if (g_cfg->call_agent_addr)
@@ -152,22 +155,27 @@ static int config_write_mgcp(struct vty *vty)
 	return CMD_SUCCESS;
 }
 
-static void dump_rtp_end(struct vty *vty, struct mgcp_rtp_state *state,
-			 struct mgcp_rtp_end *end)
+static void dump_rtp_end(struct vty *vty, struct mgcp_conn_rtp *conn)
 {
-	struct mgcp_rtp_codec *codec = &end->codec;
+	struct mgcp_rtp_state *state = &conn->state;
+	struct mgcp_rtp_end *end = &conn->end;
+	struct mgcp_rtp_codec *codec = end->codec;
+	struct rate_ctr *dropped_packets;
+
+	dropped_packets = &conn->rate_ctr_group->ctr[RTP_DROPPED_PACKETS_CTR];
 
 	vty_out(vty,
-		"   Timestamp Errs: %d->%d%s"
-		"   Dropped Packets: %d%s"
+		"   Timestamp Errs: %" PRIu64 "->%" PRIu64 "%s"
+		"   Dropped Packets: %" PRIu64 "%s"
 		"   Payload Type: %d Rate: %u Channels: %d %s"
 		"   Frame Duration: %u Frame Denominator: %u%s"
 		"   FPP: %d Packet Duration: %u%s"
 		"   FMTP-Extra: %s Audio-Name: %s Sub-Type: %s%s"
 		"   Output-Enabled: %d Force-PTIME: %d%s",
-		state->in_stream.err_ts_counter,
-		state->out_stream.err_ts_counter, VTY_NEWLINE,
-		end->dropped_packets, VTY_NEWLINE,
+		state->in_stream.err_ts_ctr->current,
+		state->out_stream.err_ts_ctr->current,
+	        VTY_NEWLINE,
+		dropped_packets->current, VTY_NEWLINE,
 		codec->payload_type, codec->rate, codec->channels, VTY_NEWLINE,
 		codec->frame_duration_num, codec->frame_duration_den,
 		VTY_NEWLINE, end->frames_per_packet, end->packet_duration_ms,
@@ -205,8 +213,7 @@ static void dump_trunk(struct vty *vty, struct mgcp_trunk_config *cfg,
 				 * connection types (E1) as soon as
 				 * the implementation is available */
 				if (conn->type == MGCP_CONN_TYPE_RTP) {
-					dump_rtp_end(vty, &conn->u.rtp.state,
-						     &conn->u.rtp.end);
+					dump_rtp_end(vty, &conn->u.rtp);
 				}
 			}
 		}
@@ -279,41 +286,70 @@ DEFUN(cfg_mgcp_bind_early,
 	return CMD_WARNING;
 }
 
-static void parse_range(struct mgcp_port_range *range, const char **argv)
-{
-	range->range_start = atoi(argv[0]);
-	range->range_end = atoi(argv[1]);
-	range->last_port = g_cfg->net_ports.range_start;
-}
-
 #define RTP_STR "RTP configuration\n"
 #define UDP_PORT_STR "UDP Port number\n"
 #define NET_START_STR "First UDP port allocated\n"
 #define RANGE_START_STR "Start of the range of ports\n"
 #define RANGE_END_STR "End of the range of ports\n"
 
-DEFUN(cfg_mgcp_rtp_net_range,
-      cfg_mgcp_rtp_net_range_cmd,
-      "rtp net-range <0-65534> <0-65534>",
+DEFUN(cfg_mgcp_rtp_port_range,
+      cfg_mgcp_rtp_port_range_cmd,
+      "rtp port-range <1024-65534> <1025-65535>",
       RTP_STR "Range of ports to use for the NET side\n"
       RANGE_START_STR RANGE_END_STR)
 {
-	parse_range(&g_cfg->net_ports, argv);
+	int start;
+	int end;
+
+	start = atoi(argv[0]);
+	end = atoi(argv[1]);
+
+	if (end < start) {
+		vty_out(vty, "range end port (%i) must be greater than the range start port (%i)!%s",
+			end, start, VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (start & 1) {
+		vty_out(vty, "range must begin at an even port number, autocorrecting port (%i) to: %i%s",
+			start, start & 0xFFFE, VTY_NEWLINE);
+		start &= 0xFFFE;
+	}
+
+	if ((end & 1) == 0) {
+		vty_out(vty, "range must end at an odd port number, autocorrecting port (%i) to: %i%s",
+			end, end | 1, VTY_NEWLINE);
+		end |= 1;
+	}
+
+	g_cfg->net_ports.range_start = start;
+	g_cfg->net_ports.range_end = end;
+	g_cfg->net_ports.last_port = g_cfg->net_ports.range_start;
+
 	return CMD_SUCCESS;
 }
+ALIAS_DEPRECATED(cfg_mgcp_rtp_port_range,
+		 cfg_mgcp_rtp_net_range_cmd,
+		 "rtp net-range <0-65534> <0-65534>",
+		 RTP_STR "Range of ports to use for the NET side\n"
+		 RANGE_START_STR RANGE_END_STR)
 
-DEFUN(cfg_mgcp_rtp_net_bind_ip,
-      cfg_mgcp_rtp_net_bind_ip_cmd,
-      "rtp net-bind-ip A.B.C.D",
+DEFUN(cfg_mgcp_rtp_bind_ip,
+      cfg_mgcp_rtp_bind_ip_cmd,
+      "rtp bind-ip A.B.C.D",
       RTP_STR "Bind endpoints facing the Network\n" "Address to bind to\n")
 {
 	osmo_talloc_replace_string(g_cfg, &g_cfg->net_ports.bind_addr, argv[0]);
 	return CMD_SUCCESS;
 }
+ALIAS_DEPRECATED(cfg_mgcp_rtp_bind_ip,
+		 cfg_mgcp_rtp_net_bind_ip_cmd,
+		 "rtp net-bind-ip A.B.C.D",
+		 RTP_STR "Bind endpoints facing the Network\n" "Address to bind to\n")
 
-DEFUN(cfg_mgcp_rtp_no_net_bind_ip,
-      cfg_mgcp_rtp_no_net_bind_ip_cmd,
-      "no rtp net-bind-ip",
+DEFUN(cfg_mgcp_rtp_no_bind_ip,
+      cfg_mgcp_rtp_no_bind_ip_cmd,
+      "no rtp bind-ip",
       NO_STR RTP_STR "Bind endpoints facing the Network\n"
       "Address to bind to\n")
 {
@@ -321,6 +357,11 @@ DEFUN(cfg_mgcp_rtp_no_net_bind_ip,
 	g_cfg->net_ports.bind_addr = NULL;
 	return CMD_SUCCESS;
 }
+ALIAS_DEPRECATED(cfg_mgcp_rtp_no_bind_ip,
+		 cfg_mgcp_rtp_no_net_bind_ip_cmd,
+		 "no rtp net-bind-ip",
+		 NO_STR RTP_STR "Bind endpoints facing the Network\n"
+		 "Address to bind to\n")
 
 DEFUN(cfg_mgcp_rtp_net_bind_ip_probing,
       cfg_mgcp_rtp_net_bind_ip_probing_cmd,
@@ -510,7 +551,7 @@ DEFUN(cfg_mgcp_number_endp,
       "Number options\n" "Endpoints available\n" "Number endpoints\n")
 {
 	/* + 1 as we start counting at one */
-	g_cfg->trunk.number_endpoints = atoi(argv[0]) + 1;
+	g_cfg->trunk.vty_number_endpoints = atoi(argv[0]) + 1;
 	return CMD_SUCCESS;
 }
 
@@ -956,7 +997,7 @@ DEFUN(tap_rtp,
 	struct mgcp_trunk_config *trunk;
 	struct mgcp_endpoint *endp;
 	struct mgcp_conn_rtp *conn;
-	uint32_t conn_id;
+        const char *conn_id = NULL;
 
 	trunk = find_trunk(g_cfg, atoi(argv[0]));
 	if (!trunk) {
@@ -980,11 +1021,11 @@ DEFUN(tap_rtp,
 
 	endp = &trunk->endpoints[endp_no];
 
-	conn_id = strtoul(argv[2], NULL, 10);
+	conn_id = argv[2];
 	conn = mgcp_conn_get_rtp(endp, conn_id);
 	if (!conn) {
-		vty_out(vty, "Conn ID %s/%d is invalid.%s",
-			argv[2], conn_id, VTY_NEWLINE);
+		vty_out(vty, "Conn ID %s is invalid.%s",
+			conn_id, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -1032,7 +1073,7 @@ DEFUN(free_endp, free_endp_cmd,
 	}
 
 	endp = &trunk->endpoints[endp_no];
-	mgcp_release_endp(endp);
+	mgcp_endp_release(endp);
 	return CMD_SUCCESS;
 }
 
@@ -1102,7 +1143,7 @@ DEFUN(cfg_mgcp_osmux,
 	 * allow to turn it on yet. */
 	vty_out(vty, "OSMUX currently unavailable in this software version.%s", VTY_NEWLINE);
 	return CMD_WARNING;
-
+#if 0
 	if (strcmp(argv[0], "on") == 0)
 		g_cfg->osmux = OSMUX_USAGE_ON;
 	else if (strcmp(argv[0], "only") == 0)
@@ -1114,6 +1155,7 @@ DEFUN(cfg_mgcp_osmux,
 	}
 
 	return CMD_SUCCESS;
+#endif
 }
 
 DEFUN(cfg_mgcp_osmux_ip,
@@ -1164,6 +1206,14 @@ DEFUN(cfg_mgcp_osmux_dummy,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_mgcp_domain,
+      cfg_mgcp_domain_cmd,
+      "domain NAME", "domain\n" "qualified domain name\n")
+{
+	osmo_strlcpy(g_cfg->domain, argv[0], sizeof(g_cfg->domain));
+	return CMD_SUCCESS;
+}
+
 int mgcp_vty_init(void)
 {
 	install_element_ve(&show_mgcp_cmd);
@@ -1181,8 +1231,11 @@ int mgcp_vty_init(void)
 	install_element(MGCP_NODE, &cfg_mgcp_bind_port_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_bind_early_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_rtp_net_range_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_rtp_port_range_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_rtp_net_bind_ip_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_rtp_bind_ip_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_rtp_no_net_bind_ip_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_rtp_no_bind_ip_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_rtp_net_bind_ip_probing_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_rtp_no_net_bind_ip_probing_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_rtp_ip_dscp_cmd);
@@ -1222,6 +1275,7 @@ int mgcp_vty_init(void)
 	install_element(MGCP_NODE, &cfg_mgcp_osmux_dummy_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_allow_transcoding_cmd);
 	install_element(MGCP_NODE, &cfg_mgcp_no_allow_transcoding_cmd);
+	install_element(MGCP_NODE, &cfg_mgcp_domain_cmd);
 
 	install_element(MGCP_NODE, &cfg_mgcp_trunk_cmd);
 	install_node(&trunk_node, config_write_trunk);
@@ -1251,18 +1305,6 @@ int mgcp_vty_init(void)
 	return 0;
 }
 
-static int allocate_trunk(struct mgcp_trunk_config *trunk)
-{
-	if (mgcp_endpoints_allocate(trunk) != 0) {
-		LOGP(DLMGCP, LOGL_ERROR,
-		     "Failed to allocate %d endpoints on trunk %d.\n",
-		     trunk->number_endpoints, trunk->trunk_nr);
-		return -1;
-	}
-
-	return 0;
-}
-
 int mgcp_parse_config(const char *config_file, struct mgcp_config *cfg,
 		      enum mgcp_role role)
 {
@@ -1286,17 +1328,18 @@ int mgcp_parse_config(const char *config_file, struct mgcp_config *cfg,
 		return -1;
 	}
 
-	if (allocate_trunk(&g_cfg->trunk) != 0) {
+	if (mgcp_endpoints_allocate(&g_cfg->trunk) != 0) {
 		LOGP(DLMGCP, LOGL_ERROR,
-		     "Failed to initialize the virtual trunk.\n");
+		     "Failed to initialize the virtual trunk (%d endpoints)\n",
+		     g_cfg->trunk.number_endpoints);
 		return -1;
 	}
 
 	llist_for_each_entry(trunk, &g_cfg->trunks, entry) {
-		if (allocate_trunk(trunk) != 0) {
+		if (mgcp_endpoints_allocate(trunk) != 0) {
 			LOGP(DLMGCP, LOGL_ERROR,
-			     "Failed to initialize E1 trunk %d.\n",
-			     trunk->trunk_nr);
+			     "Failed to initialize trunk %d (%d endpoints)\n",
+			     trunk->trunk_nr, trunk->number_endpoints);
 			return -1;
 		}
 	}
